# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'

import pandas as pd
from numpy import unique as npunique

from covid import COVID_DEATHS, COVID_CONFIRMED, COVID_RECOVERED, COUNTRY_WB_MATCH, WORLD_POPULATION
from covid.tools import check_string


def _get_cases(case_type, index_col=1):
    """

    :param case_type:
    :return:
    """
    case_type = check_string(case_type, {'deaths', 'confirmed cases', 'recovered cases'})
    if case_type == 'deaths':
        table = pd.read_csv(COVID_DEATHS, index_col=index_col)
    elif case_type == 'confirmed cases':
        table = pd.read_csv(COVID_CONFIRMED, index_col=index_col)
    else:
        table = pd.read_csv(COVID_RECOVERED, index_col=index_col)

    date = pd.DatetimeIndex(table.loc[:, "1/22/20"::].keys())

    return date, table


def get_population(country):
    """

    :param country:
    :return:
    """
    if country in "Taiwan*":
        return 23508428  # Last estimate from CIA World Factbook

    try:
        country = COUNTRY_WB_MATCH[country]
    except KeyError:
        pass
    finally:
        try:
            return WORLD_POPULATION[country]
        except KeyError:
            return None


def map_data(case_type, date):
    """

    :param case_type:
    :param date:
    :return:
    """
    time, table = _get_cases(case_type)
    countries = table.index.to_series()
    provinces = table.loc[:, "Province/State"]
    province_and_country = countries.copy()
    province_and_country[~provinces.isna()] = provinces.loc[~provinces.isna()].str.cat(
        countries[~provinces.isna()], sep=", ")

    date = f"{date.month}/{date.day}/{date.year - 2000}"
    province_state = [province for province in province_and_country]
    nb_cases = [value for value in table.loc[:, date]]
    latitude = [lat for lat in table.loc[:, "Lat"]]
    longitude = [lon for lon in table.loc[:, "Long"]]

    for country in npunique(countries):

        if province_and_country[province_and_country == country].size == 0:
            province_state.append(country)
            nb_cases.append(table.loc[[country], date].values.sum())
            latitude.append(table.loc[[country], "Lat"].values.mean())
            longitude.append(table.loc[[country], "Long"].values.mean())

    return province_state, nb_cases, latitude, longitude


def retrieve_cases(countries, case_type):
    """

    :param countries: list of (str) country names
    :param case_type: {'deaths', 'cases', 'recovered'}
    :return:
    """
    time, table = _get_cases(case_type)
    results = []

    if countries[0] == "all":
        cases = table.loc[:, "1/22/20"::].sum(axis=0)

        results.append(("Global", time, cases))

    else:
        for country in countries:
            try:
                country_cases = table.loc[[country], :]  # Using a list return a dataframe...
            except KeyError:
                country_cases = table.loc[table["Province/State"] == country]

            if country_cases.shape[0] != 1:
                if not any(country_cases["Province/State"].isna()):
                    country_cases = country_cases.loc[country, "1/22/20"::].sum(axis=0)
                else:
                    country_cases = country_cases.loc[country_cases["Province/State"].isna(), "1/22/20"::]
            else:
                country_cases = country_cases.loc[country, "1/22/20"::]

            results.append((country, time, country_cases.values.flatten()))

    return results



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from collections.abc import Collection

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


def check_type(*args) -> None:
    """Check type of arguments

    :param args: tuple list of argument/type
    :return:
    """
    if len(args) % 2 == 0:
        for item in range(0, len(args), 2):
            if not isinstance(args[item], args[item + 1]):
                raise TypeError("Type of argument {} is '{}' but must be '{}'".format(
                    item//2 + 1, type(args[item]).__name__, args[item + 1].__name__))


def check_string(string, list_of_strings):
    """ Check validity of and return string against list of valid strings

    :param string: searched string
    :param list_of_strings: list/tuple/set of valid strings string is to be checked against
    :return: validate string from list of strings if match
    """
    check_type(string, str, list_of_strings, Collection)
    [check_type(x, str) for x in list_of_strings]

    output_string = []

    for item in list_of_strings:
        if item.lower().startswith(string.lower()):
            output_string.append(item)

    if len(output_string) == 1:
        return output_string[0]
    elif len(output_string) == 0:
        raise ValueError("input must match one of those: {}".format(list_of_strings))
    elif len(output_string) > 1:
        raise ValueError("input match more than one valid value among {}".format(list_of_strings))



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'


from covid.plotting import plot_cases
from matplotlib import pyplot as plt

countries = ["Germany", "Sweden", "Netherlands", "France"]

plot_cases(countries, "deaths", doubling_factor=2, plot_type="log")
plot_cases("Sweden", "deaths", plot_type="daily")
plot_cases("all", "confirmed", doubling_factor=2, plot_type="log")

plt.show()



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

from setuptools import setup, find_packages

import covid


with open("README.md", 'r') as fh:
    long_description = fh.read()

with open("requirements.txt") as req:
    install_req = req.read().splitlines()

setup(name='pycovid-19',
      version=covid.__version__,
      description='Covid Python plots and maps',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://framagit.org/benjaminpillot/covid-19',
      author='Benjamin Pillot',
      author_email='benjaminpillot@riseup.net',
      install_requires=install_req,
      python_requires='>=3',
      license='GNU GPL v3.0',
      packages=find_packages(),
      zip_safe=False)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'
__version__ = '1.01'

import os
import tempfile
import world_bank_data as wb
from urllib.request import urlretrieve

from pandas import read_csv

covid_dir = os.path.dirname(os.path.realpath(__file__))
data_dir = os.path.join(covid_dir, "data")

COVID_DEATHS = os.path.join(tempfile.gettempdir(), "covid_deaths.csv")
COVID_CONFIRMED = os.path.join(tempfile.gettempdir(), "covid_confirmed.csv")
COVID_RECOVERED = os.path.join(tempfile.gettempdir(), "covid_recovered.csv")
WORLD_POPULATION = wb.get_series('SP.POP.TOTL', mrv=1, simplify_index=True)["Afghanistan"::]
COUNTRY_WB_MATCH = read_csv(os.path.join(data_dir, "country_wb_match.csv"), index_col=0,
                            skipinitialspace=True)["Country WB"]

urlretrieve("https://raw.githubusercontent.com/CSSEGISandData/COVID-19/master/csse_covid_19_data"
            "/csse_covid_19_time_series/time_series_covid19_deaths_global.csv", COVID_DEATHS)

urlretrieve("https://raw.githubusercontent.com/CSSEGISandData/COVID-19/master/csse_covid_19_data"
            "/csse_covid_19_time_series/time_series_covid19_confirmed_global.csv", COVID_CONFIRMED)

urlretrieve("https://raw.githubusercontent.com/CSSEGISandData/COVID-19/master/csse_covid_19_data"
            "/csse_covid_19_time_series/time_series_covid19_recovered_global.csv", COVID_RECOVERED)



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import datetime

import folium
from pandas.plotting import register_matplotlib_converters

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from covid.database import retrieve_cases, map_data, get_population

__author__ = 'Benjamin Pillot'
__copyright__ = 'Copyright 2019, Benjamin Pillot'
__email__ = 'benjaminpillot@riseup.net'

from covid.exceptions import OffsetError

from covid.tools import check_string

register_matplotlib_converters()


def _get_folium_color(case_type):
    case_type = check_string(case_type, {'deaths', 'confirmed cases', 'recovered cases'})
    if case_type == "deaths":
        return 'red'
    elif case_type == "confirmed cases":
        return 'orange'
    else:
        return "green"  # recovered


def _get_axes_and_database(countries, case_type):
    database = retrieve_cases(countries, case_type)
    if len(countries) > 1:
        nrows = int(np.ceil(len(countries) / 2))
        fig, axes = plt.subplots(nrows, 2)
        axes = axes.flatten()
    else:
        fig, axes = plt.subplots()
        axes = [axes]

    return axes, database


def get_world_map(case_type, normalized=False, date=None):
    """

    :param case_type:
    :param normalized: if True, normalized cases by country's population
    :param date: (datetime object) date for which displaying cases
    :return:
    """
    if date is None:
        date = datetime.date.today() - datetime.timedelta(days=1)

    if normalized:
        title = f"Prevalence ({case_type}) among population"
    else:
        title = f"Number of {case_type}"

    world_map = folium.Map(location=[30, 0], zoom_start=2)
    title_html = f'''<h3 align="center" style="font-size:20px"><b>{title} ({date.strftime('%Y-%m-%d')})</b></h3>'''
    world_map.get_root().html.add_child(folium.Element(title_html))

    (countries, nb_cases, latitude, longitude) = map_data(case_type, date)
    max_of_day = max(nb_cases)

    for country, case, lat, long in zip(countries, nb_cases, latitude, longitude):
        if case > 0:
            if normalized:
                population = get_population(country)
                if population:
                    folium.CircleMarker(
                        location=[lat, long],
                        radius=50 * np.sqrt(100 * case / population),
                        popup=folium.Popup(f'{country}<br>Prevalence: %.3f / 1000<br>{case_type.capitalize()}: '
                                           f'{case}' % (1000 * case/population), max_width=200),
                        fill=True,
                        color=_get_folium_color(case_type),
                        fill_color=_get_folium_color(case_type)).add_to(world_map)
            else:
                folium.CircleMarker(
                    location=[lat, long],
                    radius=20 * np.sqrt(case/max_of_day),
                    popup=folium.Popup(f'{country}<br>{case_type.capitalize()}: {case}', max_width=200),
                    fill=True,
                    color=_get_folium_color(case_type),
                    fill_color=_get_folium_color(case_type)).add_to(world_map)

    return world_map


def plot_cases(countries, case_type, doubling_factor=2, plot_type="log", normalized=False, min_number_of_cases=None,
               min_prevalence=None):
    """ Plot Covid-19 cases

    :param countries:
    :param case_type:
    :param doubling_factor:
    :param plot_type: {'daily', 'log'}
    :param normalized:
    :param min_number_of_cases: minimum number of cases condition for plotting
    :param min_prevalence:
    :return:
    """
    if isinstance(countries, str):
        countries = [countries]

    axes, database = _get_axes_and_database(countries, case_type)

    for ax, (country, dates, cases) in zip(axes, database):

        if normalized:
            ax.set_ylabel("Prevalence (\u2030)")
            population = get_population(country)
            cases = 1000 * cases / population
            if min_prevalence:
                dates = dates[cases >= min_prevalence]
                cases = cases[cases >= min_prevalence]
        else:
            ax.set_ylabel(f"Number of {case_type}")
            if min_number_of_cases:
                dates = dates[cases >= min_number_of_cases]
                cases = cases[cases >= min_number_of_cases]

        theoretical_cases = cases[0] * (2 ** (1 / doubling_factor)) ** np.arange(cases.size)

        if plot_type == 'log':
            for case in [theoretical_cases, cases]:
                ax.plot(dates, case)
            ax.legend([f"Situation where {case_type} double every %.f days" % doubling_factor, f"COVID-19 {case_type}"])
            ax.set_yscale("log")
        elif plot_type == 'daily':
            daily_cases = np.concatenate([[cases[0]], np.diff(cases)])
            ax.bar(dates, daily_cases)
            ax.legend([f"Daily increase of COVID-19 {case_type}"])

        ax.set_title(country)

        # format the ticks
        day_of_month = mdates.DayLocator(interval=max(int(cases.size/4), 1))
        ax.xaxis.set_major_locator(day_of_month)

    plt.subplots_adjust(hspace=0.35)


def plot_pace(countries, case_type, normalized=True, plot_type="log", do_offset=False, offset_value=0.0001,
              min_number_of_cases=None, min_prevalence=None):
    """ Plot case occurrence pace of given countries on the same figure

    :param countries:
    :param case_type:
    :param normalized: (bool)
    :param plot_type: {"normal", "log"}
    :param do_offset: (bool) offset at min_prevalence value
    :param offset_value: (per-thousand)
    :param min_number_of_cases:
    :param min_prevalence:
    :return:
    """
    if isinstance(countries, str):
        countries = [countries]

    fig, ax = plt.subplots()
    if plot_type == "log":
        ax.set_yscale("log")
    plt.grid(True, which='both')
    ax.set_ylabel(f"{case_type.capitalize()} (\u2030)")
    database = retrieve_cases(countries, case_type)
    num_dates = []

    for (country, dates, cases) in database:

        if normalized:
            population = get_population(country)
            cases = 1000 * cases / population
            if min_prevalence:
                dates = dates[cases >= min_prevalence]
                cases = cases[cases >= min_prevalence]
        else:
            if min_number_of_cases:
                dates = dates[cases >= min_number_of_cases]
                cases = cases[cases >= min_number_of_cases]

        if do_offset:
            try:
                offset_date = dates[np.where(cases >= offset_value)[0][0]]
            except IndexError:
                raise OffsetError("Offset value (= %.3f) is greater than the max prevalence/number of cases (= %.3f)"
                                  % (offset_value, cases[-1]))
            dates -= offset_date
            dates = dates.days
            ax.set_xlabel(f"Time (days), origin at {offset_value} (\u2030)")

        num_dates.append(cases.size)
        ax.plot(dates, cases)

    ax.legend(countries)
    if normalized:
        ax.set_title(f"Prevalence ({case_type}) over time")
    else:
        ax.set_title(f"Number of {case_type} over time")

    if not do_offset:
        day_of_month = mdates.DayLocator(interval=int(max(num_dates) / 4))
        ax.xaxis.set_major_locator(day_of_month)
